/*
 * Copyright (C) 2014-2016 Brian L. Browning
 *
 * This file is part of Beagle
 *
 * Beagle is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Beagle is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package blbutil;

import java.util.Arrays;

/**
 * <p>Interface {@code IntArray} represents an immutable {@code int[]} array.
 * </p>
 * Instances of class {@code IntArray} are required to be immutable.
 *
 * @author Brian L. Browning {@code <browning@uw.edu>}
 */
public interface IntArray {

    /**
     * Returns the number of elements in this {@code IntArray}.
     * @return the number of elements in this {@code IntArray}
     */
    int size();

    /**
     * Returns the specified array element.
     * @param index an array index
     * @return the specified array element
     * @throws IndexOutOfBoundsException if
     * {@code index < 0 ||  index >= this.size()}
     */
    int get(int index);

    /**
     * Returns the maximum element, or {@code Integer.MIN_VALUE} if
     * {@code this.size() == 0}.
     * @return the maximum element
     */
    default int max() {
        int max = Integer.MIN_VALUE;
        for (int j=0, n=size(); j<n; ++j) {
            if (get(j) > max) {
                max = get(j);
            }
        }
        return max;
    }

    /**
     * Returns the minimum element, or {@code Integer.MAX_VALUE} if
     * {@code this.size() == 0}.
     * @return the minimum element
     */
    default int min() {
        int min = Integer.MAX_VALUE;
        for (int j=0, n=size(); j<n; ++j) {
            if (get(j) < min) {
                min = get(j);
            }
        }
        return min;
    }

    /**
     * Returns a string representation of this {@code IntArray} by applying
     * {@code java.utils.Arrays.toString()} to an equivalent {@code int[]}
     * object.
     *
     * @return a string representation of this {@code IntArray}
     */
    @Override
    String toString();

    /**
     * Returns a string representation of this {@code IntArray} by applying
     * {@code java.utils.Arrays.toString()} to an equivalent {@code int[]}
     * object.
     *
     * @return a string representation of this {@code IntArray}.
     */
    default String asString() {
        int[] ia = new int[size()];
        for (int j=0; j<ia.length; ++j) {
            ia[j] = get(j);
        }
        return Arrays.toString(ia);
    }

    /**
     * Returns a new {@code IntArray} instance that has the same
     * sequence of integers as the specified array.
     * @param ia the array of integers to be copied
     * @param start the exclusive start of an interval containing the values
     * in the array
     * @param end the exclusive end of an interval containing the values
     * in the array
     * @return a new {@code IntArray} instance that has
     * the same sequence of integers as the specified array
     * @throws IllegalArgumentException if {@code start >= end}
     * @throws IllegalArgumentException if an out-of-range
     * element is detected
     * @throws NullPointerException if {@code ia == null}
     */
    static IntArray create(int[] ia, int start, int end) {
        if (start >= end) {
            throw new IllegalArgumentException(String.valueOf(start));
        }
        if (start >= 0) {
            if (end <= 128) {
                return new ByteArray(ia);
            }
            if (end <= 256) {
                return new UnsignedByteArray(ia);
            }
            else if (end <= 65536) {
                return new CharArray(ia);
            }
        }
        return new WrappedIntArray(ia);
    }

    /**
     * Returns a new {@code IntArray} instance that has the same
     * sequence of integers as the specified list.
     * @param il the list of integers to be copied
     * @param start the exclusive start of an interval containing the values
     * in the list
     * @param end the exclusive end of an interval containing the values
     * in the list
     * @return a new {@code IntArray} instance that has
     * the same sequence of integers as the specified list
     * @throws IllegalArgumentException if {@code start >= end}
     * @throws IllegalArgumentException if an out-of-range
     * element is detected
     * @throws NullPointerException if {@code il == null}
     */
    static IntArray create(IntList il, int start, int end) {
        if (start >= end) {
            throw new IllegalArgumentException(String.valueOf(start));
        }
        if (start >= 0) {
            if (end <= 128) {
                return new ByteArray(il);
            }
            if (end <= 256) {
                return new UnsignedByteArray(il);
            }
            else if (end <= 65536) {
                return new CharArray(il);
            }
        }
        return new WrappedIntArray(il);
    }
}
