/*
 * Copyright (C) 2014-2016 Brian L. Browning
 *
 * This file is part of Beagle
 *
 * Beagle is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Beagle is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package sample;

import haplotype.BitHapPair;
import haplotype.HapPair;
import haplotype.HapPairs;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Random;
import java.util.function.IntFunction;
import java.util.stream.Collectors;
import java.util.stream.IntStream;
import vcf.GT;

/**
 * <p>Class {@code InitTargHapPairs} has a static method for returning
 * initial target haplotype pairs.
 * </p>
 * <p>Instances of class {@code InitTargHapPairs} are immutable.
 * </p>
 *
 * @author Brian L. Browning {@code <browning@uw.edu>}
 */
public final class InitTargHapPairs {

    private InitTargHapPairs() {
        // private constructor to prevent instantiation
    }

    /**
     * Returns initial target haplotype pairs.  Missing alleles are sampled
     * according from the empirical allele frequency distribution, and input
     * genotypes with unknown phase will have their alleles randomly ordered
     * in the returned haplotype pairs.
     * @param gl the target genotype data
     * @param refHaps the reference haplotype pairs
     * @param minFreq a minimum allele frequency to be enforced
     * @param nCopies the number of haplotype pairs sampled for each target
     * individual
     * @param seed the seed for random number generation
     * @return initial target haplotype pairs
     * @throws IllegalArgumentException if
     * {@code minFreq <= 0.0 || minFreq >= 0.5 || Double.isNaN(minFreq) == true}
     * @throws IllegalArgumentException if {@code nCopies < 1}
     * @throws NullPointerException if {@code gl == null || refHapPairs == null}
     */
    public static List<HapPair> run(GT gl, HapPairs refHaps,
            double minFreq, int nCopies, long seed) {
        if (minFreq <= 0.0 || minFreq >= 0.5 || Double.isNaN(minFreq)) {
            throw new IllegalArgumentException(String.valueOf(minFreq));
        }
        if (nCopies < 1) {
            throw new IllegalArgumentException(String.valueOf(nCopies));
        }
        double[][] alFreq = alleleFreq(gl, refHaps, minFreq);
        IntFunction<List<HapPair>> mapper = mapper(gl, alFreq, nCopies, seed);
        return IntStream.range(0, gl.nSamples())
                .parallel()
                .mapToObj(mapper)
                .flatMap(list -> list.stream())
                .collect(Collectors.toCollection(ArrayList::new));
    }

    private static double[][] alleleFreq(GT gl, HapPairs refHaps,
            double minFreq) {
        IntFunction<double[]> freqMapper = freqMapper(gl, refHaps, minFreq);
        return IntStream.range(0, gl.nMarkers())
                .parallel()
                .mapToObj(freqMapper)
                .toArray(double[][]::new);
    }

    private static IntFunction<double[]> freqMapper(GT gl,
            HapPairs refHaps, double minFreq) {
        return (marker) -> {
            if (gl.isGTData()) {
                return alleleFreqGT(gl, refHaps, marker, minFreq);
            }
            else {
                return alleleFreqGL(gl, refHaps, marker, minFreq);
            }
        };
    }

    private static double[] alleleFreqGL(GT gl, HapPairs refHaps,
            int marker, double minFreq) {
        int nAlleles = gl.marker(marker).nAlleles();
        double[] alleleFreq = new double[nAlleles];
        double[] scaledFreq = new double[nAlleles];
        for (int s=0, n=gl.nSamples(); s<n; ++s) {
            for (int a2=0; a2<nAlleles; ++a2) {
                for (int a1=0; a1<=a2; ++a1) {
                    double like = gl.gl(marker, s, a1, a2);
                    if (a1!=a2) {
                        like = Math.max(like, gl.gl(marker, s, a2, a1));
                    }
                    scaledFreq[a2] += like;
                    scaledFreq[a1] += like;
                }
            }
            divideElementsBySum(scaledFreq);
            for (int j=0; j<scaledFreq.length; ++j) {
                alleleFreq[j] += scaledFreq[j];
                scaledFreq[j] = 0.0f;
            }
        }
        if (refHaps != null) {
            for (int h=0, n=refHaps.nHaps(); h<n; ++h) {
                ++alleleFreq[refHaps.allele(marker, h)];
            }
        }
        divideElementsBySum(alleleFreq);
        enforceMinFreq(alleleFreq, minFreq);
        return alleleFreq;
    }

    private static double[] alleleFreqGT(GT gl, HapPairs refHaps,
            int marker, double minFreq) {
        int nAlleles = gl.marker(marker).nAlleles();
        int[] cnts = new int[nAlleles];
        double[] freq = new double[nAlleles];
        for (int h=0, n=gl.nHaps(); h<n; ++h) {
            int allele = gl.allele(marker, h);
            if (allele>=0) {
                ++cnts[allele];
            }
        }
        if (refHaps != null) {
            for (int h=0, n=refHaps.nHaps(); h<n; ++h) {
                int allele = refHaps.allele(marker, h);
                ++cnts[allele];
            }
        }
        int sum = Arrays.stream(cnts).sum();
        for (int al=0; al<cnts.length; ++al) {
            freq[al] = (double) cnts[al] / sum;
        }
        enforceMinFreq(freq, minFreq);
        return freq;
    }

    private static void divideElementsBySum(double[] da) {
        double sum = Arrays.stream(da).sum();
        for (int j=0; j<da.length; ++j) {
            da[j] /= sum;
        }
    }

    private static void enforceMinFreq(double[] alleleFreq, double minFreq) {
        boolean changedFreq = false;
        for (int j=0; j<alleleFreq.length; ++j) {
            if (alleleFreq[j] < minFreq) {
                alleleFreq[j] = minFreq;
                changedFreq = true;
            }
        }
        if (changedFreq) {
            divideElementsBySum(alleleFreq);
        }
    }

    private static IntFunction<List<HapPair>> mapper(
            GT gl, double[][] alFreq, int maxCopies, long seed) {
        return (sample) -> {
            int nCopies = gl.isPhased(sample) ? 1 : maxCopies;
            int nMarkers = gl.nMarkers();
            Random rand = new Random(seed + sample);
            List<HapPair> list = new ArrayList<>(nCopies);
            int[][] hap1 = new int[nCopies][gl.nMarkers()];
            int[][] hap2 = new int[nCopies][gl.nMarkers()];
            boolean isGT = gl.isGTData();
            for (int m=0; m<nMarkers; ++m) {
                for (int c=0; c<nCopies; ++c) {
                    if (isGT) {
                        sampleGT(gl, m, sample, alFreq, hap1[c], hap2[c], rand);
                    }
                    else {
                        sampleGL(gl, m, sample, alFreq, hap1[c], hap2[c], rand);
                    }
                }
            }
            for (int c=0; c<nCopies; ++c) {
                int idIndex = gl.samples().idIndex(sample);
                list.add(new BitHapPair(gl.markers(), idIndex, hap1[c], hap2[c]));
            }
            return list;
        };
    }

    private static void sampleGT(GT gl, int m, int sample,
            double[][] alleleFreq, int[] hap1, int[]  hap2, Random rand) {
        boolean isPhased = gl.isPhased(m, sample);
        boolean swap = isPhased ? false : rand.nextBoolean();
        int a1 = swap ? gl.allele2(m, sample) : gl.allele1(m, sample);
        int a2 = swap ? gl.allele1(m, sample) : gl.allele2(m, sample);
        hap1[m] = a1<0 ? randAllele(alleleFreq[m], rand) : a1;
        hap2[m] = a2<0 ? randAllele(alleleFreq[m], rand) : a2;
    }

    private static void sampleGL(GT gl, int m, int sample,
            double[][] alleleFreq, int[] hap1, int[]  hap2, Random rand) {
        int a1 = randAllele(alleleFreq[m], rand);
        int a2 = randAllele(alleleFreq[m], rand);
        while (gl.gl(m, sample, a1, a2)==0.0) {
            a1 = randAllele(alleleFreq[m], rand);
            a2 = randAllele(alleleFreq[m], rand);
        }
        hap1[m] = a1;
        hap2[m] = a2;
    }

    private static int randAllele(double[] freq, Random rand) {
        double d = rand.nextDouble();
        double sum = 0.0;
        for (int j=0, n=freq.length; j<n; ++j) {
            sum += freq[j];
            if (sum >= d) {
                return j;
            }
        }
        return freq.length - 1;
    }
}
